import 'dart:async';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/src/builder/build_step.dart';
import 'package:source_gen/source_gen.dart';
import 'package:bloc_event_gen/bloc_event_gen.dart';

class BlocEventsGenerator extends GeneratorForAnnotation<BlocEvents> {
  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element is! ClassElement) {
      final name = element.name;
      throw InvalidGenerationSourceError(
        'Generator can only target classes. Remove the $BlocEvents annotation '
        'from `$name`.',
        element: element,
      );
    }

    StringBuffer code = StringBuffer();

    final params = annotation.objectValue.getField('events').toMapValue();
    if (params == null) {
      // Looks like generated types aren't available to us when this code is
      // executed. I haven't found a way to prioritize generators so this is a
      // workaround that works, albeit not optimal.
      throw InvalidGenerationSourceError(
        'Can\'t access $BlocEvents paramemeters. Do you possibly rely on '
        'generated types? If you do, try using a string description '
        'of the type instead. For example, for a generated type called '
        'Dog, try using \'Dog\'.',
      );
    }
    params.forEach((key, value) {
      final className = key.toStringValue();
      code.writeln('class $className extends $BlocEvent {');
      value.toMapValue().forEach((pKey, pValue) {
        var type;
        if (pValue.type.toString() == 'Type') {
          type = pValue.toTypeValue();
        } else if (pValue.type.toString() == 'String') {
          type = pValue.toStringValue();
        } else {
          throw InvalidGenerationSourceError(
              'Type of parameter must be one of Type or String.');
        }
        code.writeln('\tfinal $type ${pKey.toStringValue()};');
      });
      code.writeln(
          '${className}(${value.toMapValue().keys.map((it) => 'this.${it.toStringValue()}').join(', ')});');
      code.writeln('}');
    });
    code.writeln('');

    return code.toString();
  }
}
